import numpy as np
from datetime import datetime
from scipy import interpolate
from metpy.units import units

"""
    SOUNDERPY WRF UTILITIES

    Purpose of module:

        Parse WRF-ARW simulation output and build a SounderPy
        `clean_data` dictionary for a given latitude and longitude
        point to plot a sounding.

    Input type: netcdf4 dataset obj of a wrfout___.nc file

    (C) KYLE J GILLETT, UNIVERSITY OF NORTH DAKOTA, 2024
"""



#############################################
# FIND NEAREST VALUE IN ARRAY
#############################################

def find_nearest(array, value):
    """
    search through an array to find the index of the value nearest to a given value
    """
    array = np.asarray(array)
    nearest_idx = (np.abs(array - value)).argmin()
    return nearest_idx




#############################################
# PARSE DATASET TIME OBJECT
#############################################
def parse_time(ds):
    # convert byte data to string
    byte_data = b''.join(ds['Times'][:].data[0])
    time_str = byte_data.decode('utf-8')

    # replace underscores with spaces to match datetime format
    time_str = time_str.replace('_', ' ')

    # convert to datetime object
    time_obj = datetime.strptime(time_str, '%Y-%m-%d %H:%M:%S')

    return [str(time_obj)[0:4], str(time_obj)[5:7], str(time_obj)[8:10], str(time_obj)[11:16]]




#############################################
# COMPUTE TEMPERATURE FROM WRF OUTPUT
#############################################
def calc_temp(theta_pert, base_pres):
    '''
        Using theta perturbation from reference theta and pressure output from WRF,
        use a rearanged possion's theta eqn to find isobaric temperature
        where,
         > theta = T(P0/P)*Rd/Cp

        solve for T,
        > T = theta/(P0/P)*Rd/cp
    '''

    # CONSTANTS -------------------
    # specific heat of dry air
    cp = 1006  # J/kg/K

    # gas constant for dry air
    Rd = 287  # J/kg/K

    # reference values
    ref_theta = 300  # K
    ref_press = 100000  # Pa
    # ------------------------------

    # BUILD EQUATION --------------
    # find theta values using theta
    # perturbation from reference theta (300K)
    numerator = (theta_pert + ref_theta)  # K

    denominator = (ref_press / (base_pres)) ** (Rd / cp)  # unitless
    # -------------------------------

    return (numerator / denominator)






#############################################
# COMPUTE DEWPOINT FROM WRF OUTPUT
#############################################
def calc_dwpt(q, p):
    '''
        Using specific humidity and pressure output from WRF,

        1. Compute mixing ratio q/(1-q)
        2. Compute vapor pressure (w/(epsilon + w))*p
        3. Compute dewpoint using Td form of Clausius-Clapeyron

        Td form of Clausius-Clapetron derived from...
        Bolton, D., 1980: The computation of equivalent potential temperature.
            Mon. Wea. Rev., 108, 1046-1053, doi:10.1175/1520-0493(1980)108%3C1046:TCOEPT%3E2.0.CO;2.

    '''

    # CONSTANTS -------------------
    epsilon = 0.622
    # ------------------------------

    # compute mixing ratio
    w = (q) / (1 - q)

    # compute vapor pressure
    e = (w / (epsilon + w)) * p / 100

    # BUILD CLAUSIUS-CLAPEYRON EQUATION --------------
    numerator = (243.5 * np.log(e / 6.112))

    denominator = (17.67 - np.log(e / 6.112))
    # -------------------------------------------------

    return (numerator / denominator)






#############################################
# CREATE SMALL "AVG" BOX TO CREATE PROFILE
#############################################
def profile_location(lat_pt, lon_pt, lat_array, lon_array):
    # find latitude max and min from requested center pt
    max_lat_idx = find_nearest(lat_array, lat_pt + 0.1)
    min_lat_idx = find_nearest(lat_array, lat_pt - 0.1)

    # find longitude mx and min from requested center pt
    min_lon_idx = find_nearest(lon_array, lon_pt - 0.1)
    max_lon_idx = find_nearest(lon_array, lon_pt + 0.1)

    # return a list of min, max lats & min, max lons
    return [min_lat_idx, max_lat_idx, min_lon_idx, max_lon_idx]







#############################################
# BUILD AND RETURN 'clean_data' DICTIONARY
#############################################
def make_wrf_profile(ds, latlon, model_name='WRF-ARW', run_name='CONTROL RUN'):
    '''
        params
            > ds: netcdf file of WRF output
            > latlon: list of latitude and longitude `[45.56, -100.89]`
            > run_name: str, name of the run for title plotting 'CONTROL'
    '''

    # define latitude and longitude max/min index values for "bounding box"
    latmn, latmx, lonmn, lonmx = profile_location(latlon[0], latlon[1], ds['XLAT'][0, :, 0], ds['XLONG'][0, 0, :])

    # build vertical data dictionary
    vert_data = {
        'vert_T': np.mean(calc_temp(ds['T'][0, :, latmn:latmx, lonmn:lonmx], ds['PB'][0, :, latmn:latmx, lonmn:lonmx]),
                          axis=(1, 2)) - 273.15,
        'vert_p': np.mean((ds['PB'][:] + ds['P'][:])[0, :, latmn:latmx, lonmn:lonmx], axis=(1, 2)) / 100,
        'vert_z': np.mean((ds['PHB'][:] + ds['PH'][:])[0, 1:, latmn:latmx, lonmn:lonmx], axis=(1, 2)) / 9.81,
        'vert_u': np.mean(ds['U'][0, :, latmn:latmx, lonmn:lonmx], axis=(1, 2)) * 1.94384,
        'vert_v': np.mean(ds['V'][0, :, latmn:latmx, lonmn:lonmx], axis=(1, 2)) * 1.94384,
        'vert_Td': np.mean(calc_dwpt(ds['QVAPOR'][0, :, latmn:latmx, lonmn:lonmx],
                                     (ds['PB'][:] + ds['P'][:])[0, :, latmn:latmx, lonmn:lonmx]), axis=(1, 2))
    }

    # build sirface data dictionary
    sfc_data = {
        'sfc_T': np.mean(ds['T2'][0, latmn:latmx, lonmn:lonmx]) - 273.15,
        'sfc_p': np.mean(ds['PSFC'][0, latmn:latmx, lonmn:lonmx]) / 100,
        'sfc_z': np.mean(ds['HGT'][0, latmn:latmx, lonmn:lonmx]),
        'sfc_u': np.mean(ds['U10'][0, latmn:latmx, lonmn:lonmx]) * 1.94384,
        'sfc_v': np.mean(ds['V10'][0, latmn:latmx, lonmn:lonmx]) * 1.94384,
        'sfc_Td': np.mean(calc_dwpt(ds['Q2'][0, latmn:latmx, lonmn:lonmx], ds['PSFC'][0, latmn:latmx, lonmn:lonmx]))

    }

    # build latlon metadata dictionary
    latlon_data = {
        'data_lat': (ds['XLAT'][0, latmn:latmx, 0]),
        'data_lon': (ds['XLONG'][0, 0, lonmn:lonmx]),
        'data_latnum': (ds['XLAT'][0, latmn:latmx, 0]).shape[0],
        'data_lonnum': (ds['XLONG'][0, 0, lonmn:lonmx]).shape[0],
        'data_time': parse_time(ds)
    }

    # perform surface base-ing of isobaric values with sfc values
    sb_dict = {}
    new_keys = ['T', 'Td', 'u', 'v', 'z', 'p', ]
    sfc_keys = ['sfc_T', 'sfc_Td', 'sfc_u', 'sfc_v', 'sfc_z', 'sfc_p']
    vert_keys = ['vert_T', 'vert_Td', 'vert_u', 'vert_v', 'vert_z', 'vert_p']

    # create a dict of surface-based data
    for vert_key, sfc_key, new_key in zip(vert_keys, sfc_keys, new_keys):
        sb_dict[new_key] = np.insert(vert_data[vert_key][vert_data['vert_z'] >= sfc_data['sfc_z']], 0,
                                     sfc_data[sfc_key])
    sb_dict['z'] = sb_dict['z'] - sb_dict['z'][0]

    # interpolates data
    dz = 250
    soundingtop_hght = sb_dict['z'][-1]
    toplvl = int(soundingtop_hght / dz) * dz
    numlvls = int(toplvl / dz)
    interp_lvls = np.linspace(0, toplvl, numlvls + 1)

    # prepare new dicts
    keys = ['T', 'Td', 'u', 'v', 'z', 'p', ]
    units_list = ['degC', 'degC', 'kt', 'kt', 'm', 'hPa']
    interp_dict = {}
    zeros_dict = {}
    clean_data = {}

    surface_height = sfc_data['sfc_z']

    # create dict of clean data
    for key in keys:
        interp_dict[key] = (interpolate.interp1d(sb_dict['z'], sb_dict[key]))
        zeros_dict[key] = np.zeros((len(interp_lvls)))
        for zeros_arr in zeros_dict.values():
            zeros_dict[key][0] = sb_dict[key][0]
        for i in range(1, len(zeros_dict[key]), 1):
            zeros_dict[key][i] = interp_dict[key](dz * i)

    for i, unit, key in zip(range(0, len(units_list)), units_list, keys):
        clean_data[key] = zeros_dict[key] * units(unit)

    # add site & profile info to clean_data dict
    clean_data['site_info'] = {
        'site-id': 'no-site-id',
        'site-name': 'no-site-name',
        'site-lctn': 'no-site-location',
        'site-latlon': [latlon[0], latlon[1]],
        'site-elv': surface_height,
        'source': f'MODEL REANALYSIS',
        'model': f'{str.upper(model_name)}',
        'fcst-hour': f'{str.upper(run_name)}',
        'run-time': latlon_data['data_time'],
        'valid-time': latlon_data['data_time'],
        'box_area': f'0.1° BOX AVG'}

    # add plot title to clean_data dict
    clean_data['titles'] = {
        'top_title': f"{str.upper(model_name)} MODEL SIMULATION VERTICAL PROFILE | {str.upper(run_name)}",
        'left_title': f"VALID: {clean_data['site_info']['valid-time'][1]}/{clean_data['site_info']['valid-time'][2]}/{clean_data['site_info']['valid-time'][0]} {clean_data['site_info']['valid-time'][3]}Z",
        'right_title': f"{clean_data['site_info']['site-latlon'][0]}, {clean_data['site_info']['site-latlon'][1]} | {clean_data['site_info']['box_area']}    "
    }

    return clean_data


### IMPORT SOFTWARE ###
#########################################################################################################
import time
import sys
from urllib.request import urlopen
from urllib.error import HTTPError
import warnings
import bs4
import pandas as pd
import numpy as np
from numpy import loadtxt
import metpy.calc as mpcalc
from metpy.units import units

from .calc import *



"""
    SOUNDERPY ACARS 'GET-DATA' FUNCTIONS  

    Purpose of module: 

    House functions for loading and parsing ACARS profiles and 
    ACARS profile data. Functions here are explicitly called by
    the user.


    (C) KYLE J GILLETT, UNIVERSITY OF NORTH DAKOTA, 2024
"""






# define acars data class
class acars_data:
    """
    - NOTE: this is a Python ``Class``, not a function like the tools above.
       - This ``Class`` sets up a 'connection' to the ACARS data dataset.
       - After setting up a 'connection' to the data, you can search for available profiles using the class's function, ``.list_profiles()``
       - Then you may select one of the listed profiles and use it as an argument for the class's function, ``.get_profile()``. See below.

       :param year: observation year
       :type year: str, required
       :param month: observation month
       :type month: str, required
       :param day: observation day
       :type day: str, required
       :param hour: observation hour
       :type hour: str, required
    """

    # init
    def __init__(self, year, month, day, hour):
        self.year = year
        self.hour = hour
        self.month = month
        self.day = day
        self.hour = hour






    ### LIST PROFILES ###
    #########################################################################################################
    def list_profiles(self):
        """
        Returns
        -------

        Return a list of strings that represents ACARS profiles for a given date and hour.
        """

        st = time.time()
        print(f'> LIST ACARS PROFILES FUNCTION\n  ---------------------------------')

        # SET UP OU DIRECTORY REF
        # SEARCH FOR DIRECTORY FOR THE USER-SPECIFIED DATE
        data_dir = f'https://sharp.weather.ou.edu//soundings//acars//{self.year}//{self.month}//{self.day}//{self.hour}'

        # ACCESS THE RAW WEBSITE HTML & FIND THE ID_TIME KEYS
        print(f"> AVAILABLE ACARS PROFILES FOR {self.year}-{self.month}-{self.day} {self.hour}Z...")
        # SET UP BEAUTIFUL SOUP TO PARSE HTML
        # THIS WORKS AS A SORT OF JERRY-RIGGED WAY
        # TO REVEAL ALL AVAILABLE ACARS PROFILES
        # FOR A GIVEN DATE/TIME
        body = urlopen(data_dir).read().decode("utf-8")
        soup = bs4.BeautifulSoup(body, features="html.parser")

        # ADD PROFILES TO A LIST
        profiles_list = []
        for link in soup.select('a[href$=".txt"]'):
            profiles_list.append(link.get("href")[0:8])

        print('> COMPLETE --------')
        elapsed_time = time.time() - st
        print('> RUNTIME:', time.strftime("%H:%M:%S", time.gmtime(elapsed_time)))

        return profiles_list


    #########################################################################################################







    ### GET PROFILE DATA ###
    #########################################################################################################
    def get_profile(self, acars_profile, hush=False, clean_it=True):

        """
        Return a ``dict`` of 'cleaned up' ACARS observation profile data. Do so by selecting one of the profile string "IDs" listed by ``list_profiles()`` and pasting it as an argument in ``get_profile()``

        :param acars_profile: profile "ID"
        :type acars_profile: str, required
        :param hush: whether to 'hush' a read-out of thermodynamic and kinematic parameters when getting a data.
        :type hush: bool, optional, default is `False`
        :param clean_it: whether to return the raw_data object or a clean_data dict.
        :type clean_it: bool, optional, default is `True`
        """

        st = time.time()
        print(f'> ACARS DATA ACCESS FUNCTION\n  ---------------------------------')

        # SET UP OU DIR REF TO THE SPECIFIC PROFILE
        profile_url = f'https://sharp.weather.ou.edu//soundings//acars//{self.year}//{self.month}//{self.day}//{self.hour}//{acars_profile}.txt'

        # SEPARATE DATA BETWEEN HEADER AND ACTUAL DATA
        try:
            raw_data = loadtxt(urlopen(profile_url).readlines()[6:-1], dtype='str', comments="%", unpack=True)
            header = loadtxt(urlopen(profile_url).readlines()[0:3], dtype='str', comments="%", unpack=True)
        except HTTPError as err:
            if err.code == 404:
                sys.exit('! ERROR ! -- Invalid profile OR profile does not exist. Try again with a valid profile (ex: BNA_2320)')
            else:
                raise

        if clean_it:
            # PARSE DATE INFO FROM OU FILE
            year = f'20{header[1][0:2]}'
            month = header[1][2:4]
            day = header[1][4:6]
            hour = f'{header[1][7:9]}:{header[1][9:11]}'

            # PARSE PROFILE DATA FROM OU FILE IN DICT
            new_keys = ['p', 'z', 'T', 'Td', 'u', 'v']
            units_list = ['hPa', 'meter', 'degC', 'degC']
            clean_data = {}
            for new_key, idx, unit in zip(new_keys, range(0, 4), units_list):
                clean_data[new_key] = np.array([float(ele) for ele in [ele[0:-1] for ele in raw_data[idx]]]) * units(unit)

            clean_data['u'], clean_data['v'] = mpcalc.wind_components(
                [float(ele) for ele in [ele[0:-1] for ele in raw_data[5]]] * units.kts,
                [float(ele) for ele in [ele[0:-1] for ele in raw_data[4]]] * units.deg)

            # GET AIRPORT INFO FROM GITHUB AIRPORTS.CSV
            airports_csv = pd.read_csv(f'https://raw.githubusercontent.com/kylejgillett/sounderpy/main/src/AIRPORTS.csv',
                                       skiprows=7, skipinitialspace=True)
            where = [np.where(airports_csv['IATA'].str.contains(header[0], na=False, case=True))[0]][0][0]

            # ADD AIRPORT DATA INTO DICT
            keys = ['Name', 'City', 'Country', 'Latitude', 'Longitude', 'Altitude']
            airport_info = []
            for key in keys:
                airport_info.append(airports_csv[key][where])

            # add clean_data site_info dict
            clean_data['site_info'] = {
                'site-id': header[0],
                'site-name': airport_info[0],
                'site-lctn': airport_info[2],
                'site-latlon': [np.round(airport_info[3], 2), np.round(airport_info[4], 2)],
                'site-elv': str(int(airport_info[5])),
                'source': 'ACARS OBSERVED AIRCRAFT PROFILE',
                'model': 'no-model',
                'fcst-hour': 'no-fcst-hour',
                'run-time': ['none', 'none', 'none', 'none'],
                'valid-time': [f'20{header[1][0:2]}', header[1][2:4], header[1][4:6], f'{header[1][7:9]}:{header[1][9:11]}']
            }

            # add clean_data title dict
            clean_data['titles'] = {

                'top_title': f"ACARS AIRCRAFT OBSERVATION VERTICAL PROFILE",
                'left_title': f"VALID: {clean_data['site_info']['valid-time'][1]}-{clean_data['site_info']['valid-time'][2]}-{clean_data['site_info']['valid-time'][0]} {clean_data['site_info']['valid-time'][3]}Z",
                'right_title': f"{clean_data['site_info']['site-id']} - {clean_data['site_info']['site-name']}, {clean_data['site_info']['site-lctn']} | {clean_data['site_info']['site-latlon'][0]}, {clean_data['site_info']['site-latlon'][1]}    "
            }


            print('    > COMPLETE --------')
            elapsed_time = time.time() - st
            print('    > RUNTIME:', time.strftime("%H:%M:%S", time.gmtime(elapsed_time)))

            if not hush:
                print(
                    f"    > SUMMARY: {clean_data['site_info']['valid-time'][3]}Z Flight from {clean_data['site_info']['site-id']}, {clean_data['site_info']['site-name']} at {clean_data['site_info']['valid-time'][1]}-{clean_data['site_info']['valid-time'][2]}-{clean_data['site_info']['valid-time'][0]}-{clean_data['site_info']['valid-time'][3]}Z")
                warnings.filterwarnings("ignore")
                sounding_params(clean_data).print_vals()

            return clean_data

        else:
            return raw_data



import numpy as np
import metpy.calc as mpcalc
from metpy.units import units

"""
    SOUNDERPY CM1 UTILITIES  

    Purpose of module: 

     A collection of functions that parse and process CM1 input files
     into SounderPy `clean_data` dictionaries for easy analysis of 
     input_sounding files.


    (C) KYLE J GILLETT, UNIVERSITY OF NORTH DAKOTA, 2024, 2025
"""



#########################
# EXCEPTION CHECKS
#########################################################################
def check_sfc_hgt(full_z):
    if full_z[1] <= full_z[0]:
        raise ValueError(
            f"Invalid Surface Height Value | Check or add an elevation value in `meta_data_dict`. Value should be in meters (m).\nYour current elevation is {full_z[0]}m which is larger than the next height value of {full_z[1]}m")


def check_latlon(meta_data_dict):
    if len(meta_data_dict['latlon']) < 2:
        raise ValueError(
            f"Invalid latitude-longitude pair | Check or add a lat-lon pair to `meta_data_dict`. Values should be in a list, ex: [48.57, -100.98]")




#########################
# MAKE CM1 PROFILE
#########################################################################
def make_cm1_profile(filename, meta_data_dict):

    '''
        process a cm1 input_sounding file for sounderpy
    '''


    #############################################
    # CONSTANTS
    #############################################
    Rd = 287.05  # specific gas constant, dry air (J/kg/K)
    cp = 1005.7  # specific heat at constant pressure (J/kg/K)
    g = 9.80665  # gravitational accel. (m/s^2)
    P0 = 100000  # ref pres (hPa)

    #############################################
    # READ CM1 INPUT_SOUNDING VALUES
    #############################################
    with open(filename, 'r') as f:
        # extract surface values from the first line
        sfc_p, sfc_th, sfc_qv = map(float, f.readline().split())

        # extract vertical arrays from the remaining lines
        vert_z, vert_th, vert_qv, vert_u, vert_v = np.loadtxt(f, unpack=True)

    #############################################
    # SORT AND PREPARE DATA
    #############################################
    # determine length of vertical data arrays
    num_vals = len(vert_z) + 1

    # initialize empty arrays of len = num_vals +1 (for sfc val)
    full_z = np.zeros(num_vals, np.float32)
    full_th = np.zeros(num_vals, np.float32)
    full_qv = np.zeros(num_vals, np.float32)
    full_u = np.zeros(num_vals, np.float32)
    full_v = np.zeros(num_vals, np.float32)
    full_T = np.zeros(num_vals, np.float32)
    full_vp = np.zeros(num_vals, np.float32)
    full_Td = np.zeros(num_vals, np.float32)
    full_p = np.zeros(num_vals, np.float32)

    # add vertical array values at and beyond idx 1 in full arrays
    full_z[1:num_vals] = vert_z
    full_th[1:num_vals] = vert_th
    full_qv[1:num_vals] = vert_qv / 1000.
    full_u[1:num_vals] = vert_u
    full_v[1:num_vals] = vert_v

    # add surface values to index 0 in full arrays
    # if elev is not provided by the user, assume hydrostatic balance and
    # compute surface elevation using the hydrostatic equation
    if meta_data_dict['elev'] is None:
        elev = 1.70 * full_z[1] - full_z[2] + 0.25 * full_z[
            3]  # cm1 input has no surface hgt, use quadratic extrapolation to estimate sfc z
    else:
        elev = meta_data_dict['elev']

    full_z[0] = elev

    full_th[0] = float(sfc_th)
    full_qv[0] = float(sfc_qv) / 1000.  # convert to g/kg
    full_u[0] = 1.75 * full_u[1] - full_u[2] + 0.25 * full_u[
        3]  # cm1 input has no surface wind, use quadratic extrapolation to estimate sfc u & v
    full_v[0] = 1.75 * full_v[1] - full_v[2] + 0.25 * full_v[
        3]  # cm1 input has no surface wind, use quadratic extrapolation to estimate sfc u & v
    full_p[0] = float(sfc_p) * 100.  # convert to Pa for later calcs

    #############################################
    # COMPUTE NECESSARY DATA
    #############################################

    # convert hgt values to include elevation
    full_z[1:] = full_z[1:] + elev
    
    # compute remaining pressure values above the surface assuming hydrostatic balance
    # where dp = -rho*g*dz
    for k in np.arange(1, num_vals):
        full_p[k] = full_p[k - 1] * np.exp((g * (full_z[k - 1] - full_z[k])) / (
                    Rd * (((full_th[k] + full_th[k - 1]) / 2.) * (P0 / full_p[k - 1]) ** -(Rd / cp))))

    for k in np.arange(num_vals):
        # compute temperature from theta
        full_T[k] = mpcalc.temperature_from_potential_temperature(full_p[k] * units.Pa, full_th[k] * units.K).m
        # compute vapor pressure using the mixing ratio and pressure
        full_vp[k] = mpcalc.vapor_pressure(full_p[k] * units.Pa, full_qv[k]).m
        # compute dewpoint from vapor pressure
        full_Td[k] = mpcalc.dewpoint(full_vp[k] * units.Pa).m

    #check_sfc_hgt(full_z)
    check_latlon(meta_data_dict)

    #############################################
    # CONSTRUCT CLEAN_DATA DICTIONARY
    #############################################
    clean_data = {
        'T': np.array(full_T[0:-2] - 273.15) * units.degC,
        'Td': np.array(np.nan_to_num(full_Td, nan=-80)[0:-2]) * units.degC,
        'z': np.array(full_z[0:-2]) * units.m,
        'p': np.array(full_p[0:-2] / 100) * units.hPa,
        'u': np.array(full_u[0:-2] * 1.94384) * units.kts,
        'v': np.array(full_v[0:-2] * 1.94384) * units.kts}

    # add site & profile info to clean_data dict
    clean_data['site_info'] = {
        'site-id': 'no-site-id',
        'site-name': 'no-site-name',
        'site-lctn': 'no-site-location',
        'site-latlon': meta_data_dict['latlon'],
        'site-elv': meta_data_dict['elev'],
        'source': 'none',
        'model': 'custom',
        'fcst-hour': 'none',
        'run-time': 'none',
        'valid-time': 'none',
        'box_area': f'none'}

    # add plot title to clean_data dict
    clean_data['titles'] = {
        'top_title': meta_data_dict['top_title'],
        'left_title': meta_data_dict['left_title'],
        'right_title': f"{meta_data_dict['right_title']}    "
    }

    return clean_data


from pyproj import Geod
import numpy as np
from matplotlib.colors import ListedColormap
from datetime import datetime, timedelta
import fsspec
import importlib.resources as resources
import warnings
from siphon.catalog import TDSCatalog
import xarray as xr

import os
from contextlib import contextmanager, redirect_stderr, redirect_stdout
@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(os.devnull, 'w') as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)

with suppress_stdout_stderr():
    import pyart

"""
    SOUNDERPY RADAR UTILITY FUNCTIONS 

    Purpose of module: 

     A collection of radar-related helper functions for plotting radar data
     on SounderPY figures


    (C) KYLE J GILLETT, UNIVERSITY OF NORTH DAKOTA, 2025
"""


#########################
# RADAR DATA DATE HANDLER
#########################################################################
def define_radar_time(clean_data, radar_time):
    try:
        data_dt = datetime(int(clean_data['site_info']['valid-time'][0]), int(clean_data['site_info']['valid-time'][1]),
                           int(clean_data['site_info']['valid-time'][2]),
                           int(clean_data['site_info']['valid-time'][3][0:2]))
    except:
        warnings.warn("Sounding Date/time invalid for this profile. Defaulting to 'now' unless a specific datetime.datetime() was provided", Warning)
        data_dt = datetime.utcnow()
        pass

    # get current time
    curr_dt = datetime.utcnow()

    # if data time is after current time, default to now
    if data_dt > curr_dt:
        time = curr_dt.strftime('%H%M')
        datestr = curr_dt.strftime('%Y%m%d')
        dt_obj = curr_dt

    # if radar time is given as now, time is now
    elif radar_time == 'now':
        time = curr_dt.strftime('%H%M')
        datestr = curr_dt.strftime('%Y%m%d')
        dt_obj = curr_dt

    # if radar time is given as sounding, use the valid-time info from clean_data
    elif radar_time == 'sounding':
        if len(clean_data['site_info']['valid-time'][3]) == 2:
            time = f"{clean_data['site_info']['valid-time'][3]}00"
        else:
            time = clean_data['site_info']['valid-time'][3]
        datestr = data_dt.strftime('%Y%m%d')
        dt_obj = data_dt

    else:
        time = radar_time.strftime('%H%M')
        datestr = radar_time.strftime('%Y%m%d')
        dt_obj = radar_time


    return time, datestr, dt_obj
#########################################################################



#########################
# CREATE RADARSCOPE COLORMAP
#########################################################################
with resources.open_text("sounderpy", "rs_reflectivity.txt") as f:
    rs_data = np.loadtxt(f, skiprows=3, usecols=(1, 2, 3, 4))
rgb = rs_data[:, 1:] / 255.0
levels = rs_data[:, 0]
rs_expertreflect_cmap = ListedColormap(rgb)
#########################################################################



#########################
# DMS to decimal coords
#########################################################################
def dms_to_decimal(deg, minutes, seconds):
    return deg + minutes / 60 + seconds / 3600
#########################################################################



#########################
# PARSE DMS STRING FROM FILE
#########################################################################
def parse_dms_string(dms_str, is_lat=True):
    if is_lat:
        deg = int(dms_str[0:2])
        minutes = int(dms_str[2:4])
        seconds = int(dms_str[4:6])
    else:
        deg = int(dms_str[0:3])
        minutes = int(dms_str[3:5])
        seconds = int(dms_str[5:7])
    return dms_to_decimal(deg, minutes, seconds)
#########################################################################


#########################
# OPEN AND PARSE NEXRAD LOCS FILE
#########################################################################
def parse_nexrad_locs():
    with resources.open_text("sounderpy", "nexrad_locations.txt") as f:
        nexrad_locs = np.genfromtxt(f, dtype='str', delimiter=',', skip_header=1)
    site_ids = nexrad_locs[:, 0]
    latitudes = []
    longitudes = []
    for site in nexrad_locs[:, 1]:
        lat = parse_dms_string(site[0:6], is_lat=True)
        lon = -parse_dms_string(site[9:16], is_lat=False)
        latitudes.append(lat)
        longitudes.append(lon)

    return site_ids, latitudes, longitudes
#########################################################################



#########################
# FIND NEAREST NEXRAD SITE
#########################################################################
def find_nearest_station(site_ids, lats, lons, target_lat, target_lon, max_distance_km=400):
    g = Geod(ellps='sphere')
    distances_km = np.array([g.inv(target_lon, target_lat, lon, lat)[2] / 1000.0 for lat, lon in zip(lats, lons)])

    if np.all(np.isnan(distances_km)):
        return None

    min_dist = np.nanmin(distances_km)
    min_idx = np.nanargmin(distances_km)

    if min_dist < max_distance_km:
        site = site_ids[min_idx]
        site_lat = lats[min_idx]
        site_lon = lons[min_idx]
        # Handle special site replacement rule
        if site == 'KJAN':
            site = 'KDGX'
        return site[:], site_lat, site_lon
    return None, None, None
#########################################################################



#########################
# LOAD NEXRAD DATA FROM AWS
#########################################################################
def get_radar_data(nexrad_site, clean_data, radar_time):

    time, datestr, dt_obj = define_radar_time(clean_data, radar_time)

    fs = fsspec.filesystem('s3', anon=True)
    # try current time, then fallback to previous hour if needed
    for hour_offset in [0, -1]:
        radar_dt = dt_obj + timedelta(hours=hour_offset)
        time_candidates = [radar_dt + timedelta(minutes=delta) for delta in range(-15, 15)]

        files = []
        for t in time_candidates:
            prefix = t.strftime(f"s3://noaa-nexrad-level2/%Y/%m/%d/{nexrad_site}/{nexrad_site}%Y%m%d_%H%M")
            matches = fs.glob(prefix + "*")
            matches = [f for f in matches if not f.endswith('_MDM')]
            files.extend(matches)

        if files:
            files = sorted(files)
            file = files[int(len(files)/2)]
            break
    else:
        raise FileNotFoundError("    > RADAR ERROR: No single-site radar scan found")

    radar = pyart.io.read_nexrad_archive("s3://" + file)

    # extract timestamp from filename
    scan_time = str(files[0][48:50] + ':' + files[0][50:52])
    scan_date = str(files[0][39:43] + '-' + files[0][43:45] + '-' + files[-1][45:47])
    scan_timestamp = str(scan_date + ' | ' + scan_time)
    print(f"    > RADAR SCAN FOUND: Site: {nexrad_site} Scan: {scan_timestamp}")

    return radar, scan_timestamp, nexrad_site
#########################################################################



#########################
# LOAD RADAR MOSAIC DATA
#########################################################################
def get_radar_mosaic(clean_data, map_zoom, radar_time):
    '''
    GET RADAR MOSAIC DATA FOR SOUNDING PLOT

    - gets data from NCEI for 'sounding time' or now
    - data only goes back 1 month from current time
    '''

    # search the NCEI database for the file closest to
    # the requested radar date/time
    def find_file(target_time, file_names):
        closest_index = None
        closest_difference = float('inf')

        for idx, file_name in enumerate(file_names):
            # Extract the final component (ex: "1220") from the file name
            components = file_name.split('_')
            if len(components) > 1:
                # Remove file extension if present
                last_component = components[-1].split('.')[0]
                try:
                    file_time = int(last_component)
                    # Convert the target_time to an integer for comparison
                    target_time_int = int(target_time)
                    difference = abs(file_time - target_time_int)
                    if difference < closest_difference:
                        closest_difference = difference
                        closest_index = idx
                except ValueError:
                    pass  # Skip if the last component is not a valid integer

        return closest_index

    # GET TIME INFO -----------------------------------------------------------------------------------
    time, datestr, dt_obj = define_radar_time(clean_data, radar_time)

    # PULL SOUNDING DATA LAT/LON ----------------------------------------------------------------------
    data_lat = clean_data['site_info']['site-latlon'][0]
    data_lon = clean_data['site_info']['site-latlon'][1]

    # GET RADAR DATA ----------------------------------------------------------------------------------
    composite_url = 'https://thredds.ucar.edu/thredds/catalog/nexrad/composite/gini/dhr/1km/' + datestr +'/catalog.xml'
    try:
        # try accessing the NCEI TDS database
        composite_catalog = TDSCatalog(composite_url).datasets
        data_found = True
    except:
        # no data is available, return no data
        print('- no radar data available -')
        data_found = False
        pass

    if data_found == True:
        # if data is found, parse it for plotting.
        filename = composite_catalog[find_file(time, composite_catalog)].subset()
        composite_query = filename.query()
        composite_query.lonlat_box(north=data_lat + (map_zoom + 1),
                                   south=data_lat - (map_zoom + 1),
                                   east=data_lon + (map_zoom + 1),
                                   west=data_lon - (map_zoom + 1))
        composite_query.add_lonlat(value=True)
        composite_query.accept('netcdf4')
        composite_query.variables('Reflectivity')
        radar_data = filename.get_data(composite_query)
        radar_data = xr.open_dataset(xr.backends.NetCDF4DataStore(radar_data))

        return radar_data
#########################################################################


### IMPORT SOFTWARE ###
#########################################################################################################
# SOUNDERPY
from typing import List, Tuple, Optional

from .plot import __full_sounding, __full_hodograph, __composite_sounding, __vad_hodograph
from .calc import sounding_params
from .model_data import fetch_model
from .bufkit_data import fetch_bufkit
from .obs_data import fetch_obs
from .acars_data import *
from .wrf_utils import make_wrf_profile
from .cm1_utils import make_cm1_profile
#########################################################################################################





'''
    SOUNDERPY | Vertical Profile Data Retrieval and Analysis Tool For Python
    -------------------------------------------------------------------------
    SounderPy is an open-source atmospheric science Python package for vertical profile analysis. 
    This tool is designed to get data, ‘clean it up’ for simple use, and plot the data on advanced-sounding 
    plots. SounderPy was developed with the goal in mind to keep the code simple and efficient for users of 
    all experience levels and for reliability in all use cases. 


    THIS RELEASE
    -------
    Version: 3.0.9 | August 2025


    DOCUMENTATION
    -------
    Docs: https://kylejgillett.github.io/sounderpy/
    Code: https://github.com/kylejgillett/sounderpy
    PyPi: https://pypi.org/project/sounderpy/
    Operational Site: https://sounderpysoundings.anvil.app/

    COPYRIGHT
    ---------
    Created & maintained by Kyle J Gillett (@wxkylegillett) 2023, 2024, 2025
    
'''




citation_text = f"""
## ---------------------------------- SOUNDERPY ----------------------------------- ##
##          Vertical Profile Data Retrieval and Analysis Tool For Python            ##
##                    v3.0.9 | August 2025 | (C) Kyle J Gillett                     ##
##                 Docs: https://kylejgillett.github.io/sounderpy/                  ##
## --------------------- THANK YOU FOR USING THIS PACKAGE! ------------------------ ##
"""
print(citation_text)

#########################################################################################################




#########################################################################
##### GET DATA FUNCTIONS  #####
#########################################################################


###############
# MODEL REANALYSIS DATA
#########################################################################
def get_model_data(model: str, latlon: list, year: str, month: str, day: str, hour: str,
                   dataset: Optional[str] = None, box_avg_size: float = 0.10, hush: bool = False, clean_it: bool = True):

    r"""Get model reanalysis vertical profile data

       Return a ``dict`` of 'cleaned up' model reanalysis data from a given model, for a given location, date, and time

       :param model: the requested model to use (rap-ruc, era5, ncep)
       :type model: str, required
       :param latlon: the latitude & longitude pair for sounding ([44.92, -84.72])
       :type latlon: list, required
       :param year: valid year
       :type year: str, required
       :param month: valid month
       :type month: str, required
       :param day: valid day
       :type day: str, required
       :param hour: required, valid hour
       :type hour: str, required
       :param dataset: optional, target a specific dataset instead of searching for the first one with data.
       :type dataset: str, optional
       :param box_avg_size: optional, determine an area-averaged box size in degrees, default is 0.10 degrees.
       :type box_avg_size: int, optional
       :param hush: whether to 'hush' a read-out of thermodynamic and kinematic parameters when getting a data.
       :type hush: bool, optional, default is `False`
       :param clean_it: whether to return the raw_data object or a clean_data dict.
       :type clean_it: bool, optional, default is `True`


       :return: clean_data, a dict of ready-to-use vertical profile data including pressure, height, temperature, dewpoint, u-wind, v-wind, & model information
       :rtype: dict
    """

    data = fetch_model(model, latlon, year, month, day, hour, dataset, box_avg_size, hush, clean_it)

    return data



###############
# OBSERVED DATA
#########################################################################
def get_obs_data(station, year, month, day, hour, hush=False, clean_it=True):

    r"""Get observed vertical profile data
       Return a ``dict`` of 'cleaned up' observed profile data

       :param station: a three digit RAOB identifier (such as: 'DTX') or 11 digit IGRAv2 identifier (such as: 'GMM00010393')
       :type station: str, required
       :param year: launch year
       :type year: str, required
       :param month: launch month
       :type month: str, required
       :param day: launch day
       :type day: str, required
       :param hour: launch hour
       :type hour: str, required
       :param clean_it: whether to return the raw_data object or a clean_data dict.
       :type clean_it: bool, optional, default is `True`
       :return: clean_data, a dict of ready-to-use vertical profile data including pressure, height, temperature, dewpoint, u-wind, v-wind, & model information
       :rtype: dict
    """

    data = fetch_obs(station, year, month, day, hour, hush, clean_it)

    return data



###############
# BUFKIT DATA
#########################################################################
def get_bufkit_data(model, station, fcst_hour, run_year=None, run_month=None, run_day=None, run_hour=None,
                   hush=False, clean_it=True):

    r"""Get BUFKIT forecast model vertical profile data
       Return a ``dict`` of 'cleaned up' model forecast data from a given model, for a given BUFKIT site identifier, forecast hour, & model-run-date

       :param model: the requested model to use (such as hrrr, nam, gfs, etc)
       :type model: str, required
       :param station: a 3-4 digit BUFKIT site identifier
       :type station: str, required
       :param fcst_hour: valid forecast hour
       :type fcst_hour: int, required
       :param year: valid year
       :type year: str, required
       :param month: valid month
       :type month: str, required
       :param day: valid day
       :type day: str, required
       :param hour: valid hour
       :type hour: str, required
       :param hush: whether to 'hush' a read-out of thermodynamic and kinematic parameters when getting a data.
       :type hush: bool, optional, default is `False`
       :param clean_it: whether to return the raw_data object or a clean_data dict.
       :type clean_it: bool, optional, default is `True`


       :return: clean_data, a dict of ready-to-use vertical profile data including pressure, height, temperature, dewpoint, u-wind, v-wind, & model information
       :rtype: dict
    """

    data = fetch_bufkit(model, station, fcst_hour, run_year, run_month, run_day, run_hour,
                   hush, clean_it)

    return data



    

############
# ACARS DATA 
#########################################################################













#########################################################################
##### PLOTTING FUNCTIONS  #####
#########################################################################


############
# FULL SOUNDING
#########################################################################
def build_sounding(clean_data, style='full', color_blind=False, dark_mode=False, storm_motion='right_moving',
                   special_parcels=None, radar='mosaic', radar_time='sounding', map_zoom=2, modify_sfc=None,
                   show_theta=False, hodo_boundary=None, save=False, filename='sounderpy_sounding'):
    
    '''
    Return a full sounding plot of SounderPy data, ``plt``

    :param clean_data: the dictionary of data to be plotted (see :doc:`gettingdata`)
    :type clean_data: dict, required
    :param style: may be `simple` or `full`. Default is `full`.
    :type style: str, optional
    :param color_blind: whether to change the dewpoint trace line from green to blue for improved readability for color deficient users/readers. Default is ``False``
    :type color_blind: bool, optional
    :param dark_mode: ``True`` will invert the color scheme for a 'dark-mode' sounding. Default is ``False``.
    :type dark_mode: bool, optional
    :param storm_motion: the storm motion used for plotting and calculations. Default is 'right_moving'. Custom storm motions are accepted as a `list` of `floats` representing direction and speed. Ex: ``[270.0, 25.0]`` where '270.0' is the *direction in degrees* and '25.0' is the *speed in kts*. See the :ref:`storm_motions` section for more details.
    :type storm_motion: str or list of floats, optional
    :param special_parcels: a nested list of special parcels from the ``ecape_parcels`` library. The nested list should be a list of two lists (`[[a, b], [c, d]]`) where the first list should include 'highlight parcels' and second list should include 'background parcels'. For more details, see the :ref:`parcels_logic` section.
    :type special_parcels: nested `list` of two `lists`, optional
    :param radar: type of radar data to retrieve for radar inset, ``mosaic`` will plot a CONUS reflectivity mosaic (data only available within one month of current time), ``single`` will plot reflectivity from the nearest NEXRAD WRS-88D radar site (data availability depends on site). Default is ``mosaic``.
    :type radar: str, optional
    :param radar_time: radar mosaic data valid time. May be ``sounding`` (uses the valid time of the sounding data), or ``now`` (current time/date). **Note: radar mosaic data only goes back 1 month from current date**
    :type radar_time: str, optional, Default is ``sounding``.
    :param map_zoom: a 'zoom' level for the map inset as an `int`. **Note: Setting ``map_zoom=0`` will hide the map**
    :type map_zoom: int, optional, Default is ``2``.
    :param modify_sfc: a `dict` in the format ``{'T': 25, 'Td': 21, 'ws': 20, 'wd': 270}`` to modify the surface values of the ``clean_data`` dict. See the :ref:`modify_sfc` section for more details.
    :type modify_sfc: None or dict, optional, default is None
    :param show_theta: whether to replace the piecewise CAPE plot with theta & theta-e. Default is ``False``
    :type show_theta: bool, optional
    :param hodo_boundary: plot a "boundary" (straight line) on the hodograph axis. Do so by providing an angle and a color. The angle represents the angle between zero-degrees (north) and the upper-half of the boundary line, to the right. Multiple boundaries can be plotted. Default is ``None``
    :type hodo_boundary: dict of lists -- structured as ``{'angle':[], 'color':[]}``, optional
    :param save: whether to show the plot inline or save to a file. Default is ``False`` which displays the file inline.
    :type save: bool, optional
    :param filename: the filename by which a file should be saved to if ``save = True``. Default is `sounderpy_sounding`.
    :type filename: str, optional
    :return: plt, a SounderPy sounding built with Matplotlib, MetPy, SharpPy, & SounderPy.
    :rtype: plt
    '''
    
    print(f'> SOUNDING PLOTTER FUNCTION\n  ---------------------------------')

    plt = __full_sounding(clean_data, color_blind, dark_mode, storm_motion, special_parcels, radar, radar_time, map_zoom, modify_sfc, show_theta, hodo_boundary)
    if save:
        plt.savefig(filename, bbox_inches='tight')
    else:
        plt.show()






############
# FULL HODOGRAPH
#########################################################################
def build_hodograph(clean_data, dark_mode=False, storm_motion='right_moving', sr_hodo=False, modify_sfc=None,
                    radar='mosaic', radar_time='sounding', map_zoom=2, hodo_boundary=None, save=False, filename='sounderpy_hodograph'):

    '''
       Return a full sounding plot of SounderPy data, ``plt`` 

       :param clean_data: the dictionary of data to be plotted (see :doc:`gettingdata`)
       :type clean_data: dict, required
       :param save: whether to show the plot inline or save to a file. Default is ``False`` which displays the file inline.
       :type save: bool, optional
       :param filename: the filename by which a file should be saved to if ``save = True``. Default is `sounderpy_sounding`.
       :type filename: str, optional
       :param dark_mode: ``True`` will invert the color scheme for a 'dark-mode' sounding. Default is ``False``.
       :type dark_mode: bool, optional
       :param storm_motion: the storm motion used for plotting and calculations. Default is 'right_moving'. Custom storm motions are accepted as a `list` of `floats` representing direction and speed. Ex: ``[270.0, 25.0]`` where '270.0' is the *direction in degrees* and '25.0' is the *speed in kts*. See the :ref:`storm_motions` section for more details.
       :type storm_motion: str or list of floats, optional
       :param sr_hodo: transform the hodograph from ground relative to storm relative 
       :type sr_hodo: bool, optional, default is ``False``
       :param radar: type of radar data to retrieve for radar inset, ``mosaic`` will plot a CONUS reflectivity mosaic (data only available within one month of current time), ``single`` will plot reflectivity from the nearest NEXRAD WRS-88D radar site (data availability depends on site). Default is ``mosaic``.
       :type radar: str, optional
       :param radar_time: radar mosaic data valid time. May be ``sounding`` (uses the valid time of the sounding data), or ``now`` (current time/date). **Note: radar mosaic data only goes back 1 month from current date**
       :type radar_time: str, optional, Default is ``sounding``.
       :param map_zoom: a 'zoom' level for the map inset as an `int`. **Note: Setting ``map_zoom=0`` will hide the map**
       :type map_zoom: int, optional, Default is ``2``.
       :param modify_sfc: a `dict` in the format ``{'T': 25, 'Td': 21, 'ws': 20, 'wd': 270}`` to modify the surface values of the ``clean_data`` dict. See the :ref:`modify_sfc` section for more details.
       :type modify_sfc: None or dict, optional, default is None
       :param hodo_boundary: plot a "boundary" (straight line) on the hodograph axis. Do so by providing an angle and a color. The angle represents the angle between zero-degrees (north) and the upper-half of the boundary line, to the right. Multiple boundaries can be plotted. Default is ``None``
       :type hodo_boundary: dict of lists -- structured as ``{'angle':[], 'color':[]}``, optional
       :return: plt, a SounderPy sounding built with Matplotlib, MetPy, SharpPy, & SounderPy.
       :rtype: plt
    '''




    print(f'> HODOGRAPH PLOTTER FUNCTION --\n-------------------------------')

    plt = __full_hodograph(clean_data, dark_mode, storm_motion, sr_hodo, modify_sfc, radar, radar_time, map_zoom, hodo_boundary)
    if save:
        plt.savefig(filename, bbox_inches='tight')
    else:
        plt.show()




############
# COMPOSITE SOUNDING
#########################################################################
def build_composite(data_list, shade_between=True, cmap='viridis', colors_to_use='none', ls_to_use='none', alphas_to_use='none',
                    lw_to_use='none', dark_mode=False, save=False, filename='sounderpy_composite'):

    '''
    Return a composite sounding plot of multiple profiles, ``plt``

    :param data_list: a list of data dictionaries for each profile to be plotted
    :type data_list: list of dicts, required
    :param shade_between: Lightly shade between the dewpoint & temperature trace. In many cases, this improves readability. Default is ``True``.
    :type shade_between: bool, optional
    :param cmap: a linear colormap, may be any custom or matplotlib cmap. Default is 'viridis'. If `colors_to_use` kwarg is provided, `colors_to_use` will be used instead.
    :type cmap: `matplotlib.colors.LinearSegmentedColormap` or `str` representing the name of a matplotlib cmap, optional
    :param colors_to_use: A list of custom matplotlib color name stings. List length must match the number of profiles listed in ``data_list``. Default is 'none'.
    :type colors_to_use: list of strings, optional
    :param alphas_to_use: A list of custom alphas (0.0-1.0). List length must match the number of profiles listed in ``data_list``. Default is 'none'. Default alpha is 1.
    :type alphas_to_use: list of floats, optional
    :param ls_to_use: A list of custom matplotlib line styles. List length must match the number of profiles listed in ``data_list``. Default is 'none'. Default line style is '-'.
    :type ls_to_use: list of stings, optional
    :param lw_to_use: A list of custom line widths. List length must match the number of profiles listed in ``data_list``. Default is 'none'. Default line width is 3.
    :type lw_to_use: list of floats, optional
    :param dark_mode: ``True`` will invert the color scheme for a 'dark-mode' sounding. Default is ``False``.
    :type dark_mode: bool, optional
    :param save: whether to show the plot inline or save to a file. Default is ``False`` which displays the file inline.
    :type save: bool, optional
    :param filename: the filename by which a file should be saved to if ``save = True``. Default is `sounderpy_sounding`.
    :type filename: str, optional
    :return: plt, a SounderPy composite sounding built with Matplotlib, MetPy, SharpPy, & SounderPy.
    :rtype: plt
    '''
    
    print(f'> COMPOSITE SOUNDING FUNCTION\n  -------------------------------')

    plt =  __composite_sounding(data_list, shade_between, cmap, colors_to_use,
            ls_to_use, alphas_to_use, lw_to_use, dark_mode)
    if save:
        plt.savefig(filename, bbox_inches='tight')
    else:
        plt.show()












############
# VAD HODOGRAPH
#########################################################################
def build_vad_hodograph(vad_data, save=False, dark_mode=False, storm_motion='right_moving', sr_hodo=False,
                        filename='vad_hodograph'):
    '''
       Return a VAD hodograph plot of SounderPy VAD data, ``plt``

       :param vad_data: the dictionary of VAD data to be plotted
       :type vad_data: dict, required
       :param save: whether to show the plot inline or save to a file. Default is ``False`` which displays the file inline.
       :type save: bool, optional
       :param filename: the filename by which a file should be saved to if ``save = True``. Default is `sounderpy_sounding`.
       :type filename: str, optional
       :param dark_mode: ``True`` will invert the color scheme for a 'dark-mode' sounding. Default is ``False``.
       :type dark_mode: bool, optional
       :param storm_motion: the storm motion used for plotting and calculations. Default is 'right_moving'. Custom storm motions are accepted as a `list` of `floats` representing direction and speed. Ex: ``[270.0, 25.0]`` where '270.0' is the *direction in degrees* and '25.0' is the *speed in kts*. See the :ref:`storm_motions` section for more details.
       :type storm_motion: str or list of floats, optional
       :param sr_hodo: transform the hodograph from ground relative to storm relative
       :type sr_hodo: bool, optional, default is ``False``
       :return: plt, a SounderPy sounding built with Matplotlib, MetPy, SharpPy, & SounderPy.
       :rtype: plt
    '''

    print(f'> VAD HODOGRAPH PLOTTER FUNCTION --\n------------------------------------')

    plt =  __vad_hodograph(vad_data, dark_mode, storm_motion, sr_hodo)
    if save:
        plt.savefig(filename, bbox_inches='tight')
    else:
        plt.show()












############
# PRINT DATA TO CONSOLE
#########################################################################
def print_variables(clean_data):
    sounding_params(clean_data, storm_motion='right_moving', modify_sfc=None).print_vals()




    
    



# IMPORTS SOFTWARE
import numpy as np
from scipy.interpolate import interp1d
from metpy.units import units
from scipy.ndimage import gaussian_filter
from metpy.units import units as mpunits


"""
    SOUNDERPY MERGE & SMOOTH PROFILE FUNCTIONS

    Purpose of module: 

     A function to merge two profiles together using weighted averaging of values along a 
     common height axis to return a single profile.


    (C) KYLE J GILLETT, UNIVERSITY OF NORTH DAKOTA, 2024, 2025
"""


# PROFILE SMOOTHING SCHEME
def smooth_profile(clean_data, sigma=0.5):
    for key in clean_data.keys():
        try:
            unit = clean_data[key].units
            clean_data[key] = gaussian_filter(clean_data[key], sigma)*mpunits(str(unit))
        except:
            pass
    return clean_data



# INTERPOLATION AND AVERAGING SCHEME
def weighted_average_interp(z_a, z_b, var_a, var_b, weight_a=0.5):
    """
    Interpolates two profiles onto a common height grid and computes a weighted average profile.

    Parameters:
    - z_a: Height array for profile A.
    - z_b: Height array for profile B.
    - var_a: Variable array (e.g., temperature, pressure) for profile A.
    - var_b: Variable array for profile B.
    - weight_a: Weight for profile A in the averaging.

    Returns:
    - avg: A weighted average array of the interpolated variables.
    """

    # build a common height grid between both profiles
    z_common = np.linspace(
        max(np.min(z_a), np.min(z_b)),
        min(np.max(z_a), np.max(z_b)),
        num=100
    )

    # interpolate height vars
    interp_func_a = interp1d(z_a, var_a, bounds_error=False, fill_value='extrapolate')
    interp_func_b = interp1d(z_b, var_b, bounds_error=False, fill_value='extrapolate')

    interp_a = interp_func_a(z_common)
    interp_b = interp_func_b(z_common)

    # compute the weighted avgs
    avg = (weight_a * interp_a) + ((1 - weight_a) * interp_b)

    return avg, z_common



def merge_profiles(profile_a, profile_b, weight_a=0.5):
    """
    Averages two profiles using height coordinates.

    Parameters:
    - profile_a: Dictionary containing height and variable arrays for profile A.
    - profile_b: Dictionary containing height and variable arrays for profile B.
    - weight_a: Weight for profile A in the averaging.

    Returns:
    - averaged_profile: Dictionary with weighted-averaged values for z, T, Td, u, v, and p.
    """

    averaged_profile = {}

    unit_list = ['m', 'degC', 'degC', 'kts', 'kts']
    key_list  = ['z', 'T', 'Td', 'u', 'v']

    for key, unit in zip(key_list, unit_list):
        avg, z_common = weighted_average_interp(
            profile_a['z'], profile_b['z'],
            profile_a[key], profile_b[key],
            weight_a
        )
        averaged_profile[key] = avg * units(unit)

    interp_p_a = interp1d(profile_a['z'], profile_a['p'], bounds_error=False, fill_value='extrapolate')
    interp_p_b = interp1d(profile_b['z'], profile_b['p'], bounds_error=False, fill_value='extrapolate')
    interp_p = (weight_a * interp_p_a(z_common)) + ((1 - weight_a) * interp_p_b(z_common))
    averaged_profile['p'] = interp_p * units.hPa

    return averaged_profile
